// $Id: recfriio.cpp 5097 2008-04-27 20:18:40Z clworld $

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <fcntl.h>

#include <inttypes.h>
#include <dirent.h>
#include <errno.h>

#include <math.h>

#include <err.h>

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>

#include <glib.h>

#include <linux/usb/ch9.h>
#include <linux/usbdevice_fs.h>

#include <boost/thread/thread.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/regex.hpp>
#include <boost/format.hpp>

#include <exception>
#include <stdexcept>

#include "ringbuf.hpp"

// 受信サイズ(usbfsのバルク転送限界サイズ)
static const uint32_t TSDATASIZE = 16384;

// FIFOバッファ設定
static const int ASYNCBUFFTIME = 10; // バッファ長 = 2秒
static const int ASYNCBUFFSIZE = 0x200000 * ASYNCBUFFTIME / TSDATASIZE; // 平均16Mbpsとする

// 
static const unsigned int REQUEST_RESERVE_NUM  = 16; // 非同期リクエスト予約数
static const unsigned int REQUEST_POLLING_WAIT = 10; // 非同期リクエストポーリング間隔(ms)

static const uint16_t TARGET_ID_VENDOR  = 0x7a69; // FriioのidVendor
static const uint16_t TARGET_ID_PRODUCT = 0x0001; // FriioのidProduct

static const uint32_t TIMEOUT_RETRY_MAX      = 3;    // 初期化時にタイムアウトエラーが起こった場合のリトライ回数
static const uint32_t TIMEOUT_RETRY_INTERVAL = 1000; // ↑の間隔(ms)

static const int TARGET_INTERFACE = 0;
static const int TARGET_ENDPOINT  = 0x81;

// LED色
static const int BON_LED_PURPLE = 0; // 紫(待機中)
static const int BON_LED_GREEN  = 1; // 緑(視聴中)
static const int BON_LED_RED    = 2; // 赤(録画中)
static const int BON_LED_WHITE  = 3; // 白(エラー、送信データサンプルがない＞、＜)

typedef enum {
	FRIIO_WHITE,
	FRIIO_BLACK,
	FRIIO_TYPES_LEN,
} FriioType;

static std::string friioTypeName[FRIIO_TYPES_LEN] = { "Friio(White)", "Friio(Black)" };

//
static const float SIGNALLEVEL_RETRY_THRESHOLD = 15.0; // 開始時の信号レベル取得で信号レベルがこれ以下であれば再度信号レベルを取得する。
static const int   SIGNALLEVEL_RETRY_MAX       = 2;    // 信号レベル再取得の最大回数
static const int   SIGNALLEVEL_RETRY_INTERVAL  = 400;  // 信号レベル再取得間隔(ms)

static const char *BASE_DIR_UDEV  = "/dev/bus/usb";  // udevで作成されるUSBのディレクトリ
static const char *BASE_DIR_USBFS = "/proc/bus/usb"; // usbfsのディレクトリ

static const char *DEVICES_UDEV  = "/dev/bus/usb/devices";  // udevのdevicesファイルパス(恐らく存在しないor間違っているが念の為)
static const char *DEVICES_USBFS = "/proc/bus/usb/devices"; // usbfsのdevicesファイルパス

static int tuner_ctrl(int fd, usbdevfs_ctrltransfer *ctrl);
static void tuner_claim(int fd, unsigned int interface);
static void tuner_release(int fd, unsigned int interface);
static bool is_friio(const std::string &devfile);
static int  open_tuner();
static bool UsbProcStreamInit(int fd);
static bool UsbProcLED(int fd, uint8_t bType);
static bool UsbProcFixInitB(int fd);
static bool UsbProcFixInitA(int fd);
static bool UsbTerminate(int fd);
static bool UsbInitalize(int fd);
static void SetChannel(int fd, const uint8_t bCh);
static float GetSignalLevel(int fd);
static int tuner_readts(int fd, uint8_t *buf, int len);

// USBコマンドのタイムアウト時に投げられます。
class timeout_error : public std::runtime_error {
public:
	timeout_error(const std::string&  __arg) : std::runtime_error(__arg) {
	};
};

struct AsyncRequest {
	usbdevfs_urb urb;
	int fd;
	uint8_t buf[TSDATASIZE];
};

static void discard_async_request(AsyncRequest *r) {
	std::cout << "discards timeouted URB." << std::endl;
	int ret = ioctl(r->fd, USBDEVFS_DISCARDURB, &(r->urb));
	if (ret < 0) {
		err(1, "can't discard urb (at timeout)");
	}
}

// 非同期リクエスト発行スレッド
class PushIoThread {
public:
	PushIoThread(int fd, RingBuf<AsyncRequest> *buf) : fd(fd), buf(buf), is_cancelled(false) {
	}
	virtual ~PushIoThread() {
	}
	
	// スレッド処理本体
	void operator()() {
		while(!is_cancelled) {
			// リクエスト待ちが一杯であればREQUEST_POLLING_WAIT(ms)待ってやりなおし。
			if (REQUEST_RESERVE_NUM <= buf->getWaitCount()) {
				// リクエスト待ちが一杯です。
				usleep(REQUEST_POLLING_WAIT * 1000);
				continue;
			}
			
			// リクエスト待ちに空きがある。
			AsyncRequest *req          = buf->getPushPtr(2000, discard_async_request);
			
			memset(&(req->urb), '\0', sizeof(usbdevfs_urb));
			memset(req->buf,    '\0', TSDATASIZE);
			
			req->urb.type              = USBDEVFS_URB_TYPE_BULK;
			req->urb.endpoint          = TARGET_ENDPOINT;			
			req->urb.status            = 0;
			req->urb.buffer            = req->buf;
			req->urb.buffer_length     = TSDATASIZE;
			req->urb.actual_length     = 0;
			req->urb.number_of_packets = 0;
			req->urb.signr             = 0;
			req->urb.usercontext       = req;
			req->fd = fd;
			
			// TODO: ロックが必要か不明。
			int r = ioctl(fd, USBDEVFS_SUBMITURB, &(req->urb));
			if (r < 0) {
				err(1, "can't send urb");
			}
		}
	}
	
	// 実行停止用
	void cancel() {
		is_cancelled = true;
	}
	
protected:
	// URB送信対象fd
	int fd;
	// リングバッファ
	RingBuf<AsyncRequest> *buf;
	// 停止フラグon?
	bool is_cancelled;
};

// 非同期リクエスト受信スレッド
class PopIoThread {
public:
	PopIoThread(int fd, RingBuf<AsyncRequest> *buf) : fd(fd), buf(buf), is_cancelled(false) {
		int pipefds[2];
		int r = pipe(pipefds);
		if (r < 0) {
			err(1, "can't create pipe.");
		}
		
		event_fd_pop  = pipefds[0];
		event_fd_push = pipefds[1];
	}
	virtual ~PopIoThread() {
		if (close(event_fd_push) < 0) {
			err(1, "can't clsoe event_fd_push.");
		}
		if (close(event_fd_pop) < 0) {
			err(1, "can't clsoe event_fd_pop.");			
		}
	}
	
	// スレッド処理本体
	void operator()() {
		fd_set readfds, writefds;
		
		while(!is_cancelled) {
			FD_ZERO(&readfds);
			FD_SET(event_fd_pop, &readfds);

			FD_ZERO(&writefds);			
			FD_SET(fd, &writefds);
			
			int maxfd = MAX(fd, event_fd_pop);
			int r = select(maxfd+1, &readfds, &writefds, NULL, NULL);
			if (r < 0) {
				err(1, "select failed.");
			}
			
			if (FD_ISSET(event_fd_pop, &readfds)) {
				// cancelから起こされた？
				uint8_t dummybuf[1];
				int r = read(event_fd_pop, dummybuf, 1);
				if (r < 0) {
					err(1, "can't read from event_fd.");
				}
			}
			
			if (FD_ISSET(fd, &writefds)) {
				// usbfsから何か来た
				reapUrbs();
			}
		}
	}
	
	// 実行停止用
	void cancel() {
		is_cancelled = true;
		// selectを起こす為に書き込む
		int r = write(event_fd_push, "n", 1);
		if (r < 0) {
			err(1, "can't write to event_fd.");
		}		
	}
	
protected:
	// usbfsから完了URBを読み込む
	void reapUrbs() {
		usbdevfs_urb *urb;
		while (ioctl(fd, USBDEVFS_REAPURBNDELAY, (void *)&urb) >= 0) {
			// 読み込めた
			AsyncRequest *req = reinterpret_cast<AsyncRequest *>(urb->usercontext);
			buf->setReady(req);
		}
	}
	
	// URB受信対象fd
	int fd;
	// リングバッファ
	RingBuf<AsyncRequest> *buf;
	// 停止フラグon?
	bool is_cancelled;
	// select中断用pipe
	int event_fd_push;
	int event_fd_pop;
};

int main(int argc, char *argv[]) {
	if (argc != 4) {
		fprintf(stderr, "usage: %s channel recsec destfile\n", argv[0]);
		exit(1);
	}
	
	int chnum  = atoi(argv[1]);
	int recsec = atoi(argv[2]);
	char *destfile = argv[3];
	
	if (chnum < 13 || 62 < chnum) {
		fprintf(stderr, "channel must be (13 <= channel <= 62).\n");
		exit(1);
	}
	
	if (recsec <= 0) {
		fprintf(stderr, "recsec must be (recsec > 0).\n");
		exit(1);
	}
	
	// 録画時間の基準開始時間
	time_t time_start = time(NULL);

	int fd = open_tuner();

	FILE *dest = fopen(destfile, "w");
	if (NULL == dest) {
		fprintf(stderr, "can't open file '%s' to write.\n", destfile);
		exit(1);
	}
	
	// 
	bool retry = true;
	
	// ファイル出力までの間にTimeoutエラーが発生した場合、USBの初期化からやりなおす。
	retry = true;
	uint32_t retry_count = 0;
	while (retry) {
		retry = false;
		
		try {
			UsbInitalize(fd);
			
			SetChannel(fd, chnum);
			
			float lev_before = GetSignalLevel(fd);
			printf("Signal level: %f\n", lev_before);
			int lev_retry_count = 0;
			while (lev_before < SIGNALLEVEL_RETRY_THRESHOLD && lev_retry_count < SIGNALLEVEL_RETRY_MAX) {
				usleep(SIGNALLEVEL_RETRY_INTERVAL * 1000);
				lev_retry_count++;
				lev_before = GetSignalLevel(fd);
				printf("Signal level: %f\n", lev_before);
			}
		} catch (timeout_error &err) {
			// Timeout -> retry.
			if (TIMEOUT_RETRY_MAX <= retry_count) {
				errx(1, "Timeout error. abort.");
			}
			
			std::cout << "Timeout error. retry." << std::endl;
			retry_count++;
			close(fd); // Tunerはリトライ時開き直す。
			usleep(TIMEOUT_RETRY_INTERVAL * 1000);
			fd = open_tuner(); // Tunerはリトライ時開き直す。
			retry = true;
		}
	}
	
	printf("Output ts file.\n");
	
	timeval tv_start;
	if (gettimeofday(&tv_start, NULL) < 0) {
		err(1, "gettimeofday failed.");
	}
	
	RingBuf<AsyncRequest> buf(ASYNCBUFFSIZE);
	PushIoThread push_func(fd, &buf);
	PopIoThread  pop_func(fd,  &buf);
	boost::thread push_thread(boost::ref(push_func));
	boost::thread pop_thread(boost::ref(pop_func));
	
	while(time(NULL) <= time_start + recsec) {
		const AsyncRequest *req = buf.getPopPtr(200);
		if (NULL == req) {
			continue;
		}
		
		if (req->urb.status != 0) {
			std::cout << "UrbStatus: " << req->urb.status << std::endl;
			continue;
		}
		
		int rlen = req->urb.actual_length;
		if (0 == rlen) {
			continue;
		}
		
		fwrite(req->buf, 1, rlen, dest);
	}

	timeval tv_end;
	if (gettimeofday(&tv_end, NULL) < 0) {
		err(1, "gettimeofday failed.");
	}
	
	push_func.cancel();
	pop_func.cancel();
	push_thread.join();
	pop_thread.join();
	
	//
	fflush(dest);
	fclose(dest);
	printf("done.\n");
	
	timeval rec_time;
	timersub(&tv_end, &tv_start, &rec_time);
	printf("Rec time: %d.%06d sec.\n", (int)rec_time.tv_sec, (int)rec_time.tv_usec);
	
	try {
		float lev_after = GetSignalLevel(fd);
		printf("Signal level: %f\n", lev_after);
	
		UsbProcLED(fd, BON_LED_PURPLE);
		UsbTerminate(fd);
	
		tuner_release(fd, TARGET_INTERFACE);
	} catch (timeout_error &err) {
		std::cout << "Timeout error. ignored." << std::endl;
		// USBのタイムアウトエラーがあった場合無視して終了する。
	}
	
	int r = close(fd);
	if (-1 == r) {
		err(1, "can't close fd.");
	}
	//////////////////////////////////////////////////////////////////////////
}

// Request=0x40 RequestNo=0x01 で Length=0 のベンダリクエストを送信
void UsbSendVendorRequest(int fd, uint16_t wValue, uint16_t wIndex)
{
	uint8_t buf[] = "";
	
	usbdevfs_ctrltransfer ctrl;
	ctrl.bRequestType = USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl.bRequest     = 0x01;
	ctrl.wValue       = GUINT16_TO_LE(wValue);
	ctrl.wIndex       = GUINT16_TO_LE(wIndex);
	ctrl.wLength      = GUINT16_TO_LE(0);
	ctrl.timeout      = GUINT32_TO_LE(1000);
	ctrl.data         = buf;
	
	tuner_ctrl(fd, &ctrl);
}

int tuner_readts(int fd, uint8_t *buf, int len) {
	usbdevfs_bulktransfer bulk;
	bulk.ep      = TARGET_ENDPOINT;
	bulk.len     = len;
	bulk.timeout = 1000;
	bulk.data    = buf;
	
	int r = ioctl(fd, USBDEVFS_BULK, &bulk);
	if (-1 == r) {
		err(1, "usb bulk failed.");
	}
	return r;
}

int tuner_ctrl(int fd, usbdevfs_ctrltransfer *ctrl) {
	int r = ioctl(fd, USBDEVFS_CONTROL, ctrl);
	if (-1 == r) {
		if (ETIMEDOUT == errno) {
			throw timeout_error("tuner_ctrl: timeout.");
		} else {
			err(1, "usb ctrl failed.");
		}
	}
	return r;
}

void tuner_claim(int fd, unsigned int interface) {
	int r = ioctl(fd, USBDEVFS_CLAIMINTERFACE, &interface);
	if (-1 == r) {
		err(1, "can't claim if 0.");
	}
}

void tuner_release(int fd, unsigned int interface) {
	int r = ioctl(fd, USBDEVFS_RELEASEINTERFACE, &interface);
	if (-1 == r) {
		err(1, "can't release if 0.");
	}
}

// チャンネル設定(送信チャンネル場号はビックエンディアン)
void SetChannel(int fd, const uint8_t bCh)
{
	if(bCh < 13 || bCh > 62)
		errx(1, "invalid channel.");

	// チャンネルデータ計算
	const uint32_t wChCode = 0x0E7F + 0x002A * (uint32_t)(bCh - 13);
	
	uint8_t sbuf[2][5];
	memset(sbuf, '\0', sizeof(sbuf));
	uint8_t rbuf[2][1];
	memset(rbuf, '\0', sizeof(rbuf));
	
	// 1回目リクエスト設定	
	sbuf[0][0] = 0xC0;
	sbuf[0][1] = (uint8_t)(wChCode >> 8);
	sbuf[0][2] = (uint8_t)(wChCode & 0xFF);
	sbuf[0][3] = 0xB2;
	sbuf[0][4] = 0x08;

	// 2回目リクエスト設定
	sbuf[1][0] = 0xC0;
	sbuf[1][1] = (uint8_t)(wChCode >> 8);
	sbuf[1][2] = (uint8_t)(wChCode & 0xFF);
	sbuf[1][3] = 0x9A;
	sbuf[1][4] = 0x50;
	
	// 1
	usbdevfs_ctrltransfer ctrl1;
	ctrl1.bRequestType = USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl1.bRequest     = 0x03;
	ctrl1.wValue       = GUINT16_TO_LE(0x3000);
	ctrl1.wIndex       = GUINT16_TO_LE(0x00FE);
	ctrl1.wLength      = GUINT16_TO_LE(sizeof(sbuf[0]));
	ctrl1.timeout      = GUINT32_TO_LE(1000);
	ctrl1.data         = sbuf[0];
	
	tuner_ctrl(fd, &ctrl1);

	// 2
	usbdevfs_ctrltransfer ctrl2;
	ctrl2.bRequestType = USB_DIR_OUT|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl2.bRequest     = 0x03;
	ctrl2.wValue       = GUINT16_TO_LE(0x3000);
	ctrl2.wIndex       = GUINT16_TO_LE(0x00FE);
	ctrl2.wLength      = GUINT16_TO_LE(sizeof(sbuf[1]));
	ctrl2.timeout      = GUINT32_TO_LE(1000);
	ctrl2.data         = sbuf[1];
	
	tuner_ctrl(fd, &ctrl2);
	
	// 3
	usbdevfs_ctrltransfer ctrl3;
	ctrl3.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl3.bRequest     = 0x02;
	ctrl3.wValue       = GUINT16_TO_LE(0x3000);
	ctrl3.wIndex       = GUINT16_TO_LE(0x00B0);
	ctrl3.wLength      = GUINT16_TO_LE(sizeof(rbuf[0]));
	ctrl3.timeout      = GUINT32_TO_LE(1000);
	ctrl3.data         = rbuf[0];
	
	tuner_ctrl(fd, &ctrl3);

	// 4
	usbdevfs_ctrltransfer ctrl4;
	ctrl4.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl4.bRequest     = 0x02;
	ctrl4.wValue       = GUINT16_TO_LE(0x3000);
	ctrl4.wIndex       = GUINT16_TO_LE(0x0080);
	ctrl4.wLength      = GUINT16_TO_LE(sizeof(rbuf[1]));
	ctrl4.timeout      = GUINT32_TO_LE(1000);
	ctrl4.data         = rbuf[1];
	
	tuner_ctrl(fd, &ctrl4);
}

float GetSignalLevel(int fd) {
	float sig_level = 0;
	uint8_t sigbuf[37];
	uint32_t dwHex;
	memset(sigbuf, '\0', sizeof(sigbuf));
	
	// 1
	usbdevfs_ctrltransfer ctrl;
	ctrl.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl.bRequest     = 0x02;
	ctrl.wValue       = GUINT16_TO_LE(0x3000);
	ctrl.wIndex       = GUINT16_TO_LE(0x0089);
	ctrl.wLength      = GUINT16_TO_LE(sizeof(sigbuf));
	ctrl.timeout      = GUINT32_TO_LE(1000);
	ctrl.data         = sigbuf;
	
	tuner_ctrl(fd, &ctrl);

	dwHex = ((uint32_t)(sigbuf[2]) << 16) | ((uint32_t)(sigbuf[3]) << 8) | (uint32_t)(sigbuf[4]);
	// 信号レベル計算
	if(!dwHex) {
		sig_level = 0.0f;
	} else {
		sig_level = (float)(68.4344993f - log10((double)dwHex) * 10.0 + pow(log10((double)dwHex) - 5.045f, (double)4.0f) / 4.09f);
	}
	
	return  sig_level;
}

// 白凡黒凡判別
void scan_tuner_type(std::map<std::string, FriioType> *result) {
	boost::filesystem::path devices_file = boost::filesystem::path(DEVICES_UDEV);
	if (!boost::filesystem::exists(devices_file)) {
		devices_file = boost::filesystem::path(DEVICES_USBFS);
	}
	
	std::cout << "devices file: " << devices_file.string() << std::endl;
	
	std::ifstream in;
	in.exceptions(std::ifstream::badbit);
	in.open(devices_file.string().c_str(), std::ios::in);
	
	// hub
	std::vector<std::string> hublist;
	// hub -> Friio種別
	std::map<std::string, FriioType> hub2type;
	// hub -> devices*
	std::map<std::string, std::vector<std::string> > hub2devs;
	
	std::string bus    = "";
	std::string parent = "";
	std::string devno  = "";
	
	boost::regex regex_t("Bus=([0-9]+) .*Prnt=([0-9]+) .*Dev#= *([0-9]+)");
	boost::regex regex_p("Vendor=([A-Fa-f0-9]+) .*ProdID=([A-Fa-f0-9]+)");
	
	std::string line;
	boost::smatch match;
	while (std::getline(in, line)) {
		
		if (line.empty()) {
			continue;
		} else if (line.at(0) == 'T') {
			// トポロジ
			if (boost::regex_search(line, match, regex_t)) {
				bus    = (boost::format("%03d") % std::atoi(match.str(1).c_str())).str();
				parent = (boost::format("%03d") % std::atoi(match.str(2).c_str())).str();
				devno  = (boost::format("%03d") % std::atoi(match.str(3).c_str())).str();
			}
		} else if (line.at(0) == 'P') {
			if (boost::regex_search(line, match, regex_p)) {
				std::string vendor  = match.str(1);
				std::string product = match.str(2);
				
				std::string hub = bus + "/" + parent;
				std::string dev = bus + "/" + devno;
				
				if (vendor == "058f" && product == "9522") {
					hublist.push_back(hub);
					hub2type[hub] = FRIIO_WHITE;
				} else if (vendor == "058f" && product == "9520") {
					hublist.push_back(hub);
					hub2type[hub] = FRIIO_BLACK;
				}
				
				hub2devs[hub].push_back(dev);
			}
		}
	}
	in.close();
	
	for (std::vector<std::string>::iterator iter = hublist.begin(); iter != hublist.end(); ++iter) {
		std::string hub = *iter;
		FriioType friio_type = hub2type[hub];
		
		for (std::vector<std::string>::iterator deviter = hub2devs[hub].begin(); deviter != hub2devs[hub].end(); ++deviter) {
			(*result)[*deviter] = friio_type;
		}
	}
}

int open_tuner() {
	boost::filesystem::path base_dir = boost::filesystem::path(BASE_DIR_UDEV);
	if (!boost::filesystem::is_directory(base_dir)) {
		base_dir = boost::filesystem::path(BASE_DIR_USBFS);
	}
	
	std::cout << "Search friios from dir: " << base_dir << std::endl;
	
	FriioType target_type = FRIIO_WHITE;
	
	// 白黒判別用
	std::map<std::string, FriioType> tuner_type_map;
	scan_tuner_type(&tuner_type_map);
	
	std::vector<std::string> friios;
	// base_dirからfriioをデバイスファイルを探す。
	boost::filesystem::directory_iterator end;
	for (boost::filesystem::directory_iterator bus_iter(base_dir); bus_iter != end; ++bus_iter) {
		// バスでループ
		if (boost::filesystem::is_directory(*bus_iter)) {
			// ディレクトリのみ
			for (boost::filesystem::directory_iterator dev_iter(*bus_iter); dev_iter != end; ++dev_iter) {
				// バスに繋がってるデバイスでループ
				if (!boost::filesystem::is_directory(*dev_iter)) {
					// Friioであるかチェック
					if (is_friio(dev_iter->string())) {
						// ここまで来たらこのデバイスはfriioです。
						std::string this_device = bus_iter->leaf() + "/" + dev_iter->leaf();
						if (tuner_type_map.count(this_device) > 0) {
							if (target_type == tuner_type_map[this_device]) {
								friios.push_back(dev_iter->string());
							} else {
								std::cout << dev_iter->string() << " is "
								          << friioTypeName[tuner_type_map[this_device]] << ". skipped." << std::endl;
							}
						} else {
							std::cout << dev_iter->string() << " is Friio(Unknown). skipped." << std::endl;
						}
					}
				}
			}
		}
	}
	
	if (0 == friios.size()) {
		errx(1, "no friio found.");		
	}
	
	// 使用可能なFriioを探す
	int result_fd = -1;
	std::string result_devfile = "";
	
	std::cout << friios.size() << " friios found:" << std::endl;
	for (std::vector<std::string>::iterator iter = friios.begin(); iter != friios.end(); ++iter) {
		std::string devfile = *iter;
		
		std::cout << devfile << ": ";
		int fd = open(devfile.c_str(), O_RDWR);
		if (-1 == fd) {
			int errno_bak = errno;
			char buf[1024];
			memset(buf, '\0', sizeof(buf));
			std::cout << "open failed: " << strerror_r(errno_bak, buf, sizeof(buf)) << std::endl;
			continue;
		}
		
		if (-1 != result_fd) {
			// もう見つかっているのでclaimされているかのチェックだけ行う。
			
			usbdevfs_getdriver driver_info;
			memset(&driver_info, '\0', sizeof(driver_info));
			driver_info.interface = TARGET_INTERFACE;
			int r = ioctl(fd, USBDEVFS_GETDRIVER, &driver_info);
			if (r < 0) {
				int errno_bak = errno;
				
				if (errno_bak == ENODATA) { // 未使用の場合ここにくる。
					std::cout << "not used." << std::endl;
					close(fd);
					continue;
				}
				
				// エラー
				char buf[1024];
				memset(buf, '\0', sizeof(buf));
				std::cout << "getdriver failed: " << strerror_r(errno_bak, buf, sizeof(buf)) << std::endl;
				close(fd);
				continue;
			}
			
			std::cout << "used by " << driver_info.driver << std::endl;
			
			close(fd);
		} else {
			// まだ見つかってないのでclaimする。
			unsigned int intf = TARGET_INTERFACE;
			int r = ioctl(fd, USBDEVFS_CLAIMINTERFACE, &intf);
			if (r < 0) {
				int errno_bak = errno;				
				if (errno_bak == EBUSY) { // 使用中かチェック
					std::cout << "busy." << std::endl;
					close(fd);
					continue;
				}
			
				// エラー
				char buf[1024];
				memset(buf, '\0', sizeof(buf));
				std::cout << "claim failed: " << strerror_r(errno_bak, buf, sizeof(buf)) << std::endl;
				close(fd);
				continue;			
			}
	
			// 使用可能なfriioを発見
			result_fd = fd;
			result_devfile = devfile;
			std::cout << "use this friio." << std::endl;
		}
	}
	
	if (-1 == result_fd) {
		errx(1, "no friio can be used.");
	}
	
	std::cout << "device: " << result_devfile << std::endl;
	return result_fd;
}

bool is_friio(const std::string &devfile) {
	int f = open(devfile.c_str(), O_RDONLY);
	if (-1 == f) {
		if (errno == EACCES) {
			// 読み込み不可
			return false;
		}
		err(1, "can't open usbdevfile to read '%s'", devfile.c_str());
	}
	
	usb_device_descriptor usb_desc;
	memset(&usb_desc, '\0', sizeof(usb_desc));
	ssize_t rlen = read(f, &usb_desc, sizeof(usb_desc));
	if (-1 == rlen) {
		err(1, "can't read usbdevfile '%s'", devfile.c_str());
	}
	
	if (-1 == close(f)) {
		err(1, "can't close usbdevfile '%s'", devfile.c_str());
	}
	
	if (TARGET_ID_VENDOR == usb_desc.idVendor && TARGET_ID_PRODUCT == usb_desc.idProduct) {
		return true;
	} else {
		return false;
	}
}

// USB初期化処理
bool UsbInitalize(int fd)
{
	int iError = 0;
	bool bRet = true;
	try	{
		// コントロール固定処理Ａ
		//bRet = UsbProcFixInitA(ctrlHandle);
		if(!bRet) throw 2;
		// コントロールＬＥＤ制御(紫)
		//bRet = UsbProcLED(ctrlHandle, BON_LED_PURPLE);
		if(!bRet) throw 3;
		// コントロール固定処理Ｂ
		//bRet = UsbProcFixInitB(ctrlHandle);
		if(!bRet) throw 4;
		// コントロール固定処理Ａ
		bRet = UsbProcFixInitA(fd);
		if(!bRet) throw 6;
		// コントロール固定処理Ｂ
		bRet = UsbProcFixInitB(fd);
		if(!bRet) throw 7;
		// ストリーム制御データ
		bRet = UsbProcStreamInit(fd);
		if(!bRet) throw 8;
		// コントロールＬＥＤ制御(緑)
		bRet = UsbProcLED(fd, BON_LED_GREEN);
		if(!bRet) throw 9;
		// 初期化＼(＾o＾)／お疲れ様
	} catch(int err)
	{
		errx(1, "init failed.");
	}

	return iError==0;
}

// USB終了処理
bool UsbTerminate(int fd)
{
	return true;
}

// 固定処理Ａ
bool UsbProcFixInitA(int fd)
{
	UsbSendVendorRequest(fd, 0x0002, 0x0011);
	UsbSendVendorRequest(fd, 0x0000, 0x0011);
	
	//
	uint8_t buf1[] = "";		
	usbdevfs_ctrltransfer ctrl1;
	ctrl1.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl1.bRequest     = 0x02;
	ctrl1.wValue       = GUINT16_TO_LE(0xA000);
	ctrl1.wIndex       = GUINT16_TO_LE(0x00FE);
	ctrl1.wLength      = GUINT16_TO_LE(1);
	ctrl1.timeout      = GUINT32_TO_LE(1000);
	ctrl1.data         = buf1;
	
	tuner_ctrl(fd, &ctrl1);
	
	//
	uint8_t buf2[] = "";
	usbdevfs_ctrltransfer ctrl2;
	ctrl2.bRequestType = USB_DIR_IN|USB_TYPE_VENDOR|USB_RECIP_DEVICE;
	ctrl2.bRequest     = 0x02;
	ctrl2.wValue       = GUINT16_TO_LE(0x3000);
	ctrl2.wIndex       = GUINT16_TO_LE(0x0080);
	ctrl2.wLength      = GUINT16_TO_LE(1);
	ctrl2.timeout      = GUINT32_TO_LE(1000);
	ctrl2.data         = buf2;
	
	tuner_ctrl(fd, &ctrl2);
	
	UsbSendVendorRequest(fd, 0x0004, 0x0030);
	
	return TRUE;
}

// 固定処理Ｂ
bool UsbProcFixInitB(int fd)
{
	UsbSendVendorRequest(fd, 0x3040, 0x01);
	UsbSendVendorRequest(fd, 0x3038, 0x04);
	UsbSendVendorRequest(fd, 0x3040, 0x05);
	UsbSendVendorRequest(fd, 0x3040, 0x07);
	UsbSendVendorRequest(fd, 0x304F, 0x0F);
	UsbSendVendorRequest(fd, 0x3021, 0x11);
	UsbSendVendorRequest(fd, 0x300B, 0x12);
	UsbSendVendorRequest(fd, 0x302F, 0x13);
	UsbSendVendorRequest(fd, 0x3031, 0x14);
	UsbSendVendorRequest(fd, 0x3002, 0x16);
	UsbSendVendorRequest(fd, 0x30C4, 0x21);
	UsbSendVendorRequest(fd, 0x3020, 0x22);
	UsbSendVendorRequest(fd, 0x3079, 0x2C);
	UsbSendVendorRequest(fd, 0x3034, 0x2D);
	UsbSendVendorRequest(fd, 0x3000, 0x2F);
	UsbSendVendorRequest(fd, 0x3028, 0x30);
	UsbSendVendorRequest(fd, 0x3031, 0x31);
	UsbSendVendorRequest(fd, 0x30DF, 0x32);
	UsbSendVendorRequest(fd, 0x3001, 0x38);
	UsbSendVendorRequest(fd, 0x3078, 0x39);
	UsbSendVendorRequest(fd, 0x3033, 0x3B);
	UsbSendVendorRequest(fd, 0x3033, 0x3C);
	UsbSendVendorRequest(fd, 0x3090, 0x48);
	UsbSendVendorRequest(fd, 0x3068, 0x51);
	UsbSendVendorRequest(fd, 0x3038, 0x5E);
	UsbSendVendorRequest(fd, 0x3000, 0x71);
	UsbSendVendorRequest(fd, 0x3008, 0x72);
	UsbSendVendorRequest(fd, 0x3000, 0x77);
	UsbSendVendorRequest(fd, 0x3021, 0xC0);
	UsbSendVendorRequest(fd, 0x3010, 0xC1);
	UsbSendVendorRequest(fd, 0x301A, 0xE4);
	UsbSendVendorRequest(fd, 0x301F, 0xEA);
	UsbSendVendorRequest(fd, 0x3000, 0x77);
	UsbSendVendorRequest(fd, 0x3000, 0x71);
	UsbSendVendorRequest(fd, 0x3000, 0x71);
	UsbSendVendorRequest(fd, 0x300C, 0x76);
	return true;
}

bool UsbProcLED(int fd, uint8_t bType)
{
	UsbSendVendorRequest(fd, 0x000F, 0x06);
	switch(bType)
	{
	case BON_LED_PURPLE:
		{
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x000B,0x00);
			UsbSendVendorRequest(fd, 0x000F,0x00);

			UsbSendVendorRequest(fd, 0x0003,0x00);
			UsbSendVendorRequest(fd, 0x0007,0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0009, 0x00);
			UsbSendVendorRequest(fd, 0x000D, 0x00);
		}
		break;
	case BON_LED_GREEN:
		{
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0001, 0x00);
			UsbSendVendorRequest(fd, 0x0005, 0x00);
		}
		break;
	case BON_LED_RED:
		{
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);

			UsbSendVendorRequest(fd, 0x000B, 0x00);
			UsbSendVendorRequest(fd, 0x000F, 0x00);

			UsbSendVendorRequest(fd, 0x0003, 0x00);
			UsbSendVendorRequest(fd, 0x0007, 0x00);
			UsbSendVendorRequest(fd, 0x0001, 0x00);
			UsbSendVendorRequest(fd, 0x0005, 0x00);
		}
		break;
	case BON_LED_WHITE:
		{
			// 白点滅が故意に起こせなかったのでパターン情報ないです。
			// EP0へのコントロール転送で
			// 　・開始(40 01 0F 00 06 00 00 00)
			// 　・終了(40 01 00 00 06 00 00 00)
			// なデータの羅列がLEDコントロールぽいので、解った方追記希望
		}
		break;
	}
	UsbSendVendorRequest(fd, 0x0000, 0x06);
	return true;
}

// ストリーム初期化処理
bool UsbProcStreamInit(int fd)
{
	UsbSendVendorRequest(fd, 0x08, 0x33);
	UsbSendVendorRequest(fd, 0x40, 0x37);
	UsbSendVendorRequest(fd, 0x1F, 0x3A);
	UsbSendVendorRequest(fd, 0xFF, 0x3B);
	UsbSendVendorRequest(fd, 0x1F, 0x3C);
	UsbSendVendorRequest(fd, 0xFF, 0x3D);
	UsbSendVendorRequest(fd, 0x00, 0x38);
	UsbSendVendorRequest(fd, 0x00, 0x35);
	UsbSendVendorRequest(fd, 0x00, 0x39);
	UsbSendVendorRequest(fd, 0x00, 0x36);
	return true;
}
